@@ -25,11 +25,11 @@
 
 #include "kerncompat.h"
 #include "kernel-lib/radix-tree.h"
-#include "ctree.h"
-#include "disk-io.h"
-#include "print-tree.h"
-#include "transaction.h"
-#include "volumes.h"
+#include "kernel-shared/ctree.h"
+#include "kernel-shared/disk-io.h"
+#include "kernel-shared/print-tree.h"
+#include "kernel-shared/transaction.h"
+#include "kernel-shared/volumes.h"
 #include "cmds/commands.h"
 #include "common/utils.h"
 #include "common/help.h"
@@ -46,7 +46,7 @@ static void print_extents(struct extent_buffer *eb)
 		return;
 
 	if (btrfs_is_leaf(eb)) {
-		btrfs_print_leaf(eb);
+		btrfs_print_leaf(eb, BTRFS_PRINT_TREE_DEFAULT);
 		return;
 	}
 
@@ -207,6 +207,8 @@ static const char * const cmd_inspect_dump_tree_usage[] = {
 	"--bfs                  breadth-first traversal of the trees, print nodes, then leaves (default)",
 	"--dfs                  depth-first traversal of the trees",
 	"--hide-names           hide filenames/subvolume/xattrs and other name references",
+	"--csum-headers         print node checksums stored in headers (metadata)",
+	"--csum-items           print checksums stored in checksum items (data)",
 	NULL
 };
 
@@ -250,7 +252,7 @@ static int dump_add_tree_block(struct cache_tree *tree, u64 bytenr)
  * error.
  */
 static int dump_print_tree_blocks(struct btrfs_fs_info *fs_info,
-				  struct cache_tree *tree, bool follow)
+				  struct cache_tree *tree, unsigned int mode)
 {
 	struct cache_extent *ce;
 	struct extent_buffer *eb;
@@ -280,7 +282,7 @@ static int dump_print_tree_blocks(struct btrfs_fs_info *fs_info,
 			ret = -EIO;
 			goto next;
 		}
-		btrfs_print_tree(eb, follow, BTRFS_PRINT_TREE_DEFAULT);
+		btrfs_print_tree(eb, mode);
 		free_extent_buffer(eb);
 next:
 		remove_cache_extent(tree, ce);
@@ -302,6 +304,7 @@ static int cmd_inspect_dump_tree(const struct cmd_struct *cmd,
 	struct btrfs_disk_key disk_key;
 	struct btrfs_key found_key;
 	struct cache_tree block_root;	/* for multiple --block parameters */
+	struct open_ctree_flags ocf = { 0 };
 	char uuidbuf[BTRFS_UUID_UNPARSED_SIZE];
 	int ret = 0;
 	int slot;
@@ -316,7 +319,9 @@ static int cmd_inspect_dump_tree(const struct cmd_struct *cmd,
 	u64 block_bytenr;
 	struct btrfs_root *tree_root_scan;
 	u64 tree_id = 0;
-	bool follow = false;
+	unsigned int follow = 0;
+	unsigned int csum_mode = 0;
+	unsigned int print_mode;
 
 	/*
 	 * For debug-tree, we care nothing about extent tree (it's just backref
@@ -331,7 +336,9 @@ static int cmd_inspect_dump_tree(const struct cmd_struct *cmd,
 	while (1) {
 		int c;
 		enum { GETOPT_VAL_FOLLOW = 256, GETOPT_VAL_DFS, GETOPT_VAL_BFS,
-		       GETOPT_VAL_NOSCAN, GETOPT_VAL_HIDE_NAMES };
+		       GETOPT_VAL_NOSCAN, GETOPT_VAL_HIDE_NAMES,
+		       GETOPT_VAL_CSUM_HEADERS, GETOPT_VAL_CSUM_ITEMS,
+		};
 		static const struct option long_options[] = {
 			{ "extents", no_argument, NULL, 'e'},
 			{ "device", no_argument, NULL, 'd'},
@@ -345,6 +352,8 @@ static int cmd_inspect_dump_tree(const struct cmd_struct *cmd,
 			{ "dfs", no_argument, NULL, GETOPT_VAL_DFS },
 			{ "noscan", no_argument, NULL, GETOPT_VAL_NOSCAN },
 			{ "hide-names", no_argument, NULL, GETOPT_VAL_HIDE_NAMES },
+			{ "csum-headers", no_argument, NULL, GETOPT_VAL_CSUM_HEADERS },
+			{ "csum-items", no_argument, NULL, GETOPT_VAL_CSUM_ITEMS },
 			{ NULL, 0, NULL, 0 }
 		};
 
@@ -401,7 +410,7 @@ static int cmd_inspect_dump_tree(const struct cmd_struct *cmd,
 			break;
 			}
 		case GETOPT_VAL_FOLLOW:
-			follow = true;
+			follow = BTRFS_PRINT_TREE_FOLLOW;
 			break;
 		case GETOPT_VAL_DFS:
 			traverse = BTRFS_PRINT_TREE_DFS;
@@ -415,6 +424,12 @@ static int cmd_inspect_dump_tree(const struct cmd_struct *cmd,
 		case GETOPT_VAL_HIDE_NAMES:
 			open_ctree_flags |= OPEN_CTREE_HIDE_NAMES;
 			break;
+		case GETOPT_VAL_CSUM_HEADERS:
+			csum_mode |= BTRFS_PRINT_TREE_CSUM_HEADERS;
+			break;
+		case GETOPT_VAL_CSUM_ITEMS:
+			csum_mode |= BTRFS_PRINT_TREE_CSUM_ITEMS;
+			break;
 		default:
 			usage_unknown_option(cmd, argv);
 		}
@@ -459,15 +474,19 @@ static int cmd_inspect_dump_tree(const struct cmd_struct *cmd,
 
 	printf("%s\n", PACKAGE_STRING);
 
-	info = open_ctree_fs_info(argv[optind], 0, 0, 0, open_ctree_flags);
+	ocf.filename = argv[optind];
+	ocf.flags = open_ctree_flags;
+	info = open_ctree_fs_info(&ocf);
 	if (!info) {
 		error("unable to open %s", argv[optind]);
 		goto out;
 	}
 
+	print_mode = follow | traverse | csum_mode;
+
 	if (!cache_tree_empty(&block_root)) {
 		root = info->chunk_root;
-		ret = dump_print_tree_blocks(info, &block_root, follow);
+		ret = dump_print_tree_blocks(info, &block_root, print_mode);
 		goto close_root;
 	}
 
@@ -493,20 +512,20 @@ static int cmd_inspect_dump_tree(const struct cmd_struct *cmd,
 		} else {
 			if (info->tree_root->node) {
 				printf("root tree\n");
-				btrfs_print_tree(info->tree_root->node, true,
-						 traverse);
+				btrfs_print_tree(info->tree_root->node,
+					BTRFS_PRINT_TREE_FOLLOW | print_mode);
 			}
 
 			if (info->chunk_root->node) {
 				printf("chunk tree\n");
-				btrfs_print_tree(info->chunk_root->node, true,
-						 traverse);
+				btrfs_print_tree(info->chunk_root->node,
+					BTRFS_PRINT_TREE_FOLLOW | print_mode);
 			}
 
 			if (info->log_root_tree) {
 				printf("log root tree\n");
 				btrfs_print_tree(info->log_root_tree->node,
-						 true, traverse);
+					BTRFS_PRINT_TREE_FOLLOW | print_mode);
 			}
 		}
 	}
@@ -526,7 +545,8 @@ again:
 			goto close_root;
 		}
 		printf("root tree\n");
-		btrfs_print_tree(info->tree_root->node, true, traverse);
+		btrfs_print_tree(info->tree_root->node,
+					BTRFS_PRINT_TREE_FOLLOW | print_mode);
 		goto close_root;
 	}
 
@@ -536,7 +556,8 @@ again:
 			goto close_root;
 		}
 		printf("chunk tree\n");
-		btrfs_print_tree(info->chunk_root->node, true, traverse);
+		btrfs_print_tree(info->chunk_root->node,
+					BTRFS_PRINT_TREE_FOLLOW | print_mode);
 		goto close_root;
 	}
 
@@ -546,7 +567,8 @@ again:
 			goto close_root;
 		}
 		printf("log root tree\n");
-		btrfs_print_tree(info->log_root_tree->node, true, traverse);
+		btrfs_print_tree(info->log_root_tree->node,
+					BTRFS_PRINT_TREE_FOLLOW | print_mode);
 		goto close_root;
 	}
 
@@ -692,7 +714,8 @@ again:
 					       btrfs_header_level(buf));
 				} else {
 					printf(" \n");
-					btrfs_print_tree(buf, true, traverse);
+					btrfs_print_tree(buf,
+						BTRFS_PRINT_TREE_FOLLOW | print_mode);
 				}
 			}
 			free_extent_buffer(buf);
