@@ -1,4 +1,19 @@
 // SPDX-License-Identifier: GPL-2.0
+/*
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public
+ * License v2 as published by the Free Software Foundation.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public
+ * License along with this program; if not, write to the
+ * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
+ * Boston, MA 021110-1307, USA.
+ */
 
 #include <sys/ioctl.h>
 #include <unistd.h>
@@ -489,8 +504,14 @@ size_t btrfs_sb_io(int fd, void *buf, off_t offset, int rw)
 	if ((stat_buf.st_mode & S_IFMT) == S_IFBLK) {
 		ret = ioctl(fd, BLKGETZONESZ, &zone_size_sector);
 		if (ret < 0) {
-			if (errno == ENOTTY) {
-				/* No kernel support, assuming non-zoned device */
+			if (errno == ENOTTY || errno == EINVAL) {
+				/*
+				 * No kernel support, assuming non-zoned device.
+				 *
+				 * Note: older kernels before 5.11 could return
+				 * EINVAL in case the ioctl is not available,
+				 * which is wrong.
+				 */
 				zone_size_sector = 0;
 			} else {
 				error("zoned: ioctl BLKGETZONESZ failed: %m");
@@ -533,7 +554,11 @@ size_t btrfs_sb_io(int fd, void *buf, off_t offset, int rw)
 
 	ret = ioctl(fd, BLKREPORTZONE, rep);
 	if (ret) {
-		if (errno == ENOTTY) {
+		if (errno == ENOTTY || errno == EINVAL) {
+			/*
+			 * Note: older kernels before 5.11 could return EINVAL
+			 * in case the ioctl is not available, which is wrong.
+			 */
 			error("zoned: BLKREPORTZONE failed but BLKGETZONESZ works: %m");
 			exit(1);
 		}
